import grpc
from si_tests import logger
from si_tests.clients.iam import keycloak_client
from third_party import iam_api_v1_pb2
from third_party import iam_api_v1_pb2_grpc
from si_tests import settings

LOG = logger.logger


class IAMClient(object):

    def __init__(self, keycloak_ip, iam_ip_port,
                 cacert_path=settings.IAM_CACERT):

        self.keycloak_ip = keycloak_ip
        self.iam_url = iam_ip_port
        self.cacert_path = cacert_path

    def connect(self, username, password):
        self.username = username
        self.password = password

        self.client = keycloak_client.KeycloakUserClient(self.keycloak_ip,
                                                         self.username,
                                                         self.password)

        cred = grpc.ssl_channel_credentials(
            open(self.cacert_path, 'rb').read()
        )
        cred2 = grpc.access_token_call_credentials(self.client.get_token())
        composite_credentials = grpc.composite_channel_credentials(cred,
                                                                   cred2)
        channel = grpc.secure_channel(self.iam_url, composite_credentials)

        # create a stub (client)
        self.stub = iam_api_v1_pb2_grpc.IAMStub(channel)
        return self

    def get_token(self):
        return self.client.get_token()

    def get_version(self):
        empty = iam_api_v1_pb2.Empty()
        response = self.stub.GetVersions(empty)
        return response

    def list_my_grants(self):
        return self.list_grants(self.username)

    def list_users(self, filter=None):
        req = iam_api_v1_pb2.ListUserRequest(params=filter)
        user_list = []
        for user in self.stub.ListUsers(req):
            user_list.append(user)
        return user_list

    def get_this_user(self):
        return self.get_user(self.username)

    def get_user(self, username):
        req = iam_api_v1_pb2.UserRequest(username=username)
        return self.stub.ShowUser(req)

    def list_grants(self, username):
        req = iam_api_v1_pb2.UserRequest(username=username)
        grant_list = []
        for grant in self.stub.ListGrantsForUser(req):
            grant_list.append(grant)
        return grant_list

    def give_grants(self, username, role, scope):
        grant = iam_api_v1_pb2.Grant(role=role, scope=scope)
        req = iam_api_v1_pb2.GrantRequest(username=username, grant=grant)
        self.stub.GrantAccess(req)

    def revoke_grants(self, username, role, scope):
        grant = iam_api_v1_pb2.Grant(role=role, scope=scope)
        req = iam_api_v1_pb2.GrantRequest(username=username, grant=grant)
        self.stub.RevokeAccess(req)

    def list_roles_for_scope(self, scope):
        req = iam_api_v1_pb2.ScopeRequest(name=scope)
        roles_list = []
        for role in self.stub.ListRolesForScope(req):
            roles_list.append(role)
        return roles_list

    def list_scopes(self, prefix=''):
        req = iam_api_v1_pb2.ListScopeRequest(prefix=prefix)
        scope_list = []
        for role in self.stub.ListScopes(req):
            scope_list.append(role)
        return scope_list

    def list_users_for_scope(self, scope):
        req = iam_api_v1_pb2.ScopeRequest(name=scope)
        user_list = []
        for user in self.stub.ListUsersForScope(req):
            user_list.append(user)
        return user_list

    def has_grants(self, username, check_grants):
        actual_grants = self.list_grants(username)
        missing_grants = []
        same = False
        no_matches = False
        for grant in check_grants:
            gr = iam_api_v1_pb2.Grant(role=grant["role"],
                                      scope=grant["scope"])
            if gr not in actual_grants:
                missing_grants.append(gr)
            else:
                actual_grants.remove(gr)

        if not missing_grants and not actual_grants:
            same = True
        if len(check_grants) == len(missing_grants):
            no_matches = True
        return {"missing_grants": missing_grants,
                "extra_grants": actual_grants,
                "same": same,
                "no_matches": no_matches}

    def list_servicetoken(self, all=True):
        req = iam_api_v1_pb2.ListServiceTokensRequest(all=all)
        servicetoken_list = []
        for servicetoken in self.stub.ListServiceTokens(req):
            servicetoken_list.append(servicetoken)
        return servicetoken_list

    def create_servicetoken(self, alias, service, grants):
        req = iam_api_v1_pb2.CreateServiceTokenRequest(alias=alias,
                                                       service=service,
                                                       grants=grants)
        return self.stub.CreateServiceToken(req)

    def delete_servicetoken(self, id):
        req = iam_api_v1_pb2.ServiceTokenRequest(id=id)
        self.stub.DeleteServiceToken(req)

    def get_servicetoken(self, id):
        req = iam_api_v1_pb2.ServiceTokenRequest(id=id)
        return self.stub.ShowServiceToken(req)

    def get_servicetoken_by_alias(self, alias):
        for st in self.list_servicetoken():
            if st.alias == alias:
                return st
