from packaging import version
from si_tests import logger
from si_tests import settings

from si_tests.managers import bootstrap_manager


LOG = logger.logger


def get_kaas_binary_name(kaas_version):
    """
    Returns expected name of binary based on kaas release version.
    :param kaas_version: Kaas release version
    :return: Name of kaas binary
    """
    if version.parse(kaas_version) > version.parse("2.11"):
        return 'container-cloud'
    return 'kaas'


class IntrospectBootstrap(object):
    def __init__(self, openstack_client, JIRA="INTROSPECT-2021"):
        self.kubeconfig_path = None
        self.bootstrap_manager = bootstrap_manager.BootstrapManager()
        self.openstack_client = openstack_client
        self.__remote = None
        self.PREF = f"=== {JIRA}: "
        self.kind_container_name = "clusterapi-control-plane"

        self.bastion_ip = None
        self.private_key_path = None
        self.mgmt_ips = None

    @property
    def remote(self):
        if self.__remote:
            return self.__remote

        LOG.debug(f"{self.PREF} Check if the seed node alive")
        try:
            if self.bootstrap_manager.get_seed_ip(self.openstack_client):
                remote = self.bootstrap_manager.remote_seed(
                    self.openstack_client)
                remote.check_call('hostname -s')
            else:
                raise Exception("Seed node IP address not found")
        except Exception as e:
            LOG.error(f"{self.PREF} Seed node is not available via SSH: {e}")
            raise
        self.__remote = remote
        return self.__remote

    @property
    def r_base(self):
        remote_base = './'
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            kaas_version = self.bootstrap_manager.get_kaas_version()
        else:
            kaas_version = self.bootstrap_manager.get_remote_kaas_version(
                self.remote, self.openstack_client)
        kaas_binary_name = get_kaas_binary_name(kaas_version)
        if self.remote.isfile(f"bootstrap/dev/{kaas_binary_name}"):
            remote_base = 'bootstrap/dev/'
        return remote_base

    @property
    def is_kind_running(self):
        try:
            res = self.remote.execute(
                f'docker ps|grep {self.kind_container_name}')
            if len(res.stdout) > 0:
                LOG.debug(f"{self.PREF} Docker kind cluster is running")
                return True
        except Exception:
            pass
        LOG.debug(f"{self.PREF} Docker kind cluster is not running")
        return False

    def get_kubeconfig_path(self):
        kubeconfig_path = "{0}kubeconfig".format(self.r_base)
        assert self.remote.isfile(kubeconfig_path), (
            f"{self.PREF} Kubeconfig not found on {kubeconfig_path}")
        return kubeconfig_path

    def get_private_key_path(self):
        key_path = self.bootstrap_manager.get_bootstrap_ssh_keyfile_path(
                self.remote)
        assert self.remote.isfile(key_path), (
            f"{self.PREF} SSH private key not found on {key_path}")
        return key_path

    def cmd_run_in_kind(self, cmd):
        res = self.remote.check_call(
            "docker exec clusterapi-control-plane /bin/bash -c \""
            "export KUBECONFIG=/etc/kubernetes/admin.conf && " + cmd + " \"")
        return res

    def cmd_run_in_mgmt(self, cmd):
        kubeconfig_path = self.get_kubeconfig_path()
        res = self.remote.check_call(
            f"PATH=${{PATH}}:~/kaas-bootstrap/bin:"
            f"/home/ubuntu/bootstrap/dev/bin;"
            f"bash -c \"export KUBECONFIG={kubeconfig_path} && " + cmd + " \"")
        return res

    def get_mgmt_ips(self):
        """Return list of IP addresses of lcmmachines"""

        cmd = ("kubectl get lcmmachines -o=jsonpath='{.items[*].status."
               "addresses[?(@.type==\\\"InternalIP\\\")].address}'")
        mgmt_ips = []

        if self.is_kind_running:
            # Try to get lcmmachines from kind cluster first
            res = self.cmd_run_in_kind(cmd)
            mgmt_ips = res.stdout_str.split()

        if not mgmt_ips:
            # Get lcmmachines from management nodes
            res = self.cmd_run_in_mgmt(cmd)
            mgmt_ips = res.stdout_str.split()

        assert len(mgmt_ips) > 0, (
            "LCMMachines were not found on kind/mgmt clusters")
        return mgmt_ips

    def get_bastion_ip(self):
        """Return the bastion IP address"""
        cmd = ("kubectl get cluster -o=jsonpath="
               "'{.items[*].status.providerStatus.bastion.publicIP}'")
        bastion_ip = ""

        if self.is_kind_running:
            # Try to get lcmmachines from kind cluster first
            res = self.cmd_run_in_kind(cmd)
            bastion_ip = res.stdout_str

        if not bastion_ip:
            # Get lcmmachines from management nodes
            res = self.cmd_run_in_mgmt(cmd)
            bastion_ip = res.stdout_str

        assert bastion_ip, (
            "Bastion IP was not found on kind/mgmt clusters")
        return bastion_ip

    def get_ssh_cmd(self, mgmt_ip, private_key_path, bastion_ip=""):
        if bastion_ip:
            return (
                f"ssh -i {private_key_path} mcc-user@{mgmt_ip} "
                f"-oStrictHostKeyChecking=no -o proxycommand='ssh -W %h:%p "
                f"-i {private_key_path} mcc-user@{bastion_ip} "
                f"-oStrictHostKeyChecking=no' ")
        else:
            return (
                f"ssh -i {private_key_path} mcc-user@{mgmt_ip} "
                f"-oStrictHostKeyChecking=no ")

    def get_openstack_instance_details(self, ssh_cmd, mgmt_ip=""):
        """Get instance details from OpenStack"""
        details = {}

        LOG.debug(f"{self.PREF} Check the mgmt node availability)")
        try:
            res = self.remote.check_call(f"{ssh_cmd} 'hostname -s'")
            hostname = res.stdout_str
            details['hostname'] = hostname
        except Exception as e:
            LOG.error(f"{self.PREF} MGMT node '{mgmt_ip}' is not available"
                      f" via SSH: {e}")
            return details

        LOG.debug(f"{self.PREF} Get the mgmt node hypervisor name")
        try:
            server = self.openstack_client.get_server_by_name(
                hostname, verbose=False)._info
            hypervisor_hostname = server["OS-EXT-SRV-ATTR:hypervisor_hostname"]
            instance_name = server["OS-EXT-SRV-ATTR:instance_name"]
            instance_id = server["id"]
            details['os'] = {
                'hypervisor_hostname': hypervisor_hostname,
                'instance_name': instance_name,
                'instance_id': instance_id,
            }
        except Exception as e:
            LOG.error(f"{self.PREF} Unable to retrive the MGMT node "
                      f"'{hostname}' details from OpenStack: {e}")

        return details

    def init(self):
        if not self.bastion_ip:
            self.bastion_ip = self.get_bastion_ip()
        if not self.private_key_path:
            self.private_key_path = self.get_private_key_path()
        if not self.mgmt_ips:
            self.mgmt_ips = self.get_mgmt_ips()
