#    Copyright 2021 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import pytest

from si_tests import settings
from si_tests import logger

from si_tests.managers import bootstrap_manager


LOG = logger.logger


class TestProvisionAzureMgmCluster():

    @pytest.mark.usefixtures("introspect_MKE_7914")
    @pytest.mark.usefixtures("introspect_PRODX_9238")
    def test_bootstrap_kaas_on_azure(self, openstack_client, show_step):
        """Bootstrap KaaS on OpenStack VMs

        Scenario:
            1. Erase previous KaaS environment and seed node, if exists
            2. Create a seed VM using heat template
            3. Prepare seed node to run KaaS bootstrap on Azure
            4. Check Azure bootstrap templates for management cluster
            5. Bootstrap KaaS seed on OS and managed on Azure clusters
            6. Create default MCC users, if requested
            7. Check versions
            8. Wait for all pods to be Running and Ready
            9. Check that all expected pods exist
            10. Download bootstrap artifacts
            11. Erase created KaaS environment
        """

        bootstrap = bootstrap_manager.BootstrapManager()

        if settings.KEEP_ENV_BEFORE:
            LOG.info("Skip erase due KEEP_ENV_BEFORE flag is setted")
        else:
            if openstack_client._current_stack:
                # Erase the stack with the seed node
                show_step(1)
                bootstrap.step_001_erase_env_before(openstack_client)
            else:
                LOG.warning("Can't erase environment due "
                            "seed node stack is absent")

        # Create a stack with a seed node
        show_step(2)
        bootstrap.step_002_provision_seed_vm(openstack_client)

        show_step(3)
        bootstrap.step_003_prepare_seed_node(openstack_client)

        if settings.SKIP_BOOTSTRAP_TEMPLATES_CHECK:
            LOG.warning("Skip bootstrap templates check due "
                        "SKIP_BOOTSTRAP_TEMPLATES_CHECK flag is set")
            if not settings.KAAS_CUSTOM_LMA_LOGGING_ENABLED:
                bootstrap.disable_lma_logging(openstack_client)
        else:
            show_step(4)
            bootstrap.check_templates(openstack_client)

        show_step(5)
        bootstrap.step_004_deploy_kaas_cluster(openstack_client)

        # TODO continue checks from here
        if settings.CREATE_DEFAULT_MCC_USERS_VIA_CLI:
            show_step(6)
            bootstrap.step_004_create_default_mcc_users(openstack_client)

        if settings.MCC_SCHEDULED_UPGRADES:
            LOG.info("Update mcc-upgrade crd for feature upgrade scheduling")
            bootstrap.step_006_update_mcc_upgrade_crd()

        show_step(7)
        bootstrap.step_005_check_versions(openstack_client)

        show_step(8)
        bootstrap.step_006_wait_for_pods()

        show_step(9)
        bootstrap.step_006_postdeployment_checks(openstack_client)

        show_step(10)
        bootstrap.step_007_download_bootstrap_artifacts(openstack_client)

        if settings.KEEP_ENV_AFTER:
            LOG.warning("Skip erase due KEEP_ENV_AFTER flag is set")
        else:
            # Erase kaas environment after bootstrap is passed
            show_step(11)
            bootstrap.step_007_erase_env_after(openstack_client)
