import pytest
from kubernetes.client.rest import ApiException

from si_tests import logger
from si_tests.deployments.utils import wait_utils

LOG = logger.logger


@pytest.fixture(scope="class")
def setup(request, os_manager):
    request.cls.tf_helmbundle = \
        os_manager.get_tf_helmbundle(request.cls.helmbundle_name)
    request.cls.tf_deployment = os_manager.get_tfoperator_deployment()
    request.cls.tfoperator_cr = os_manager.get_tfoperator_cr()
    request.cls.api_admission = os_manager.api.api_admission


@pytest.mark.usefixtures("setup")
class TestTFAdmissionController(object):
    """Test TFOperator admission controller"""
    webhook_name = "tungstenfabric-operator-validating-webhook-configuration"
    helmbundle_name = "tungstenfabric-operator"
    tf_helmbundle = None
    tf_deployment = None
    tfoperator_cr = None
    api_admission = None

    @pytest.fixture()
    def restore_helmbundle(self):
        # Backup helmbundle
        self.spec_backup = self.tf_helmbundle.data['spec']
        yield
        # Restore helmbundle
        self.tf_helmbundle.patch({"spec": self.spec_backup})

    @pytest.fixture()
    def restore_cr(self):
        # Backup tfoperator CR specs
        self.cr_backup = self.tfoperator_cr.data['spec']
        yield
        # Restore tfoperator CR specs
        self.tfoperator_cr.patch({"spec": self.cr_backup})

    def set_admission_ctl(self, helmbundle, enabled=True):
        spec = helmbundle.data['spec']
        chart_name = 'tungstenfabric-operator'
        for chart in spec['releases']:
            if chart['name'] == chart_name:
                chart['values']['admission'] = {'enabled': enabled}
                return spec
        raise Exception("Chart '{}' not found.".format(chart_name))

    def check_deployment(self, gen):
        return self.tf_deployment.read().status.observed_generation > gen

    def check_webhook(self):
        try:
            self.api_admission.read_validating_webhook_configuration(
                self.webhook_name)
            return True
        except ApiException as e:
            if e.status == 404:
                return False

    def check_webhook_enabled(self):
        LOG.info('Validating Webhook {} is created'.format(
            self.webhook_name))
        return self.check_webhook()

    def check_webhook_disabled(self):
        LOG.info('Validating Webhook {} is removed'.format(
            self.webhook_name))
        return not self.check_webhook()

    @pytest.mark.usefixtures("restore_cr")
    def test_invalidation_request(self, show_step):
        """Check valid and non-valid requests.

            Scenario:
                1. Check validation webhook
                2. Send and check invalid request
        """

        show_step(1)
        assert self.check_webhook_enabled(), \
            "Validation Admission Webhook isn't created"

        show_step(2)
        invalid_settings = {
            "settings": {
                "tfImage": {
                    "tfImageTag": "5.1.20220101011200"
                },
                "tfVersion": "2011",
            }
        }
        with pytest.raises(ApiException) as exc_info:
            self.tfoperator_cr.patch({"spec": invalid_settings})
        assert exc_info.value.status == 403
        assert exc_info.value.reason == 'Forbidden'

    @pytest.mark.usefixtures("restore_helmbundle")
    def test_manage_admission_controller(self, os_manager, show_step):
        """Enable/disable tf-operator admission controller via helmbundle.

            Scenario:
                1. Disable admission controller
                2. Check Validating Webhook Configuration was removed
                3. Enable admission controller
                4. Check Validating Webhook Configuration was created
        """

        wait = wait_utils.Waiter(os_manager, 300, os_manager.tf_namespace)

        show_step(1)
        gen = self.tf_deployment.read().status.observed_generation
        spec = self.set_admission_ctl(self.tf_helmbundle, enabled=False)
        self.tf_helmbundle.patch({"spec": spec})

        # Check deployment has been updated
        show_step(2)
        # Skip deployment check if no changes
        if not spec == self.spec_backup:
            wait.wait(self.check_deployment, gen)
        wait.wait(self.check_webhook_disabled)

        show_step(3)
        gen = self.tf_deployment.read().status.observed_generation
        spec = self.set_admission_ctl(self.tf_helmbundle, enabled=True)
        self.tf_helmbundle.patch({"spec": spec})

        # Check deployment has been updated
        show_step(4)
        wait.wait(self.check_deployment, gen)
        wait.wait(self.check_webhook_enabled)
