import yaml
from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters, templates, utils

LOG = logger.logger

cluster_name = settings.TARGET_CLUSTER
namespace_name = settings.TARGET_NAMESPACE
region = settings.TARGET_REGION


def test_add_bm_control_node(kaas_manager):
    """Add one more control and worker node."""

    # Get namespace
    LOG.info("Namespace name - %s", namespace_name)
    ns = kaas_manager.get_namespace(namespace_name)

    # Create ssh key
    public_key_name = "{cluster_name}-key".format(
        cluster_name=cluster_name)
    LOG.info("Public key name - {public_key_name}".format(
        public_key_name=public_key_name))
    ns.get_publickey(public_key_name)

    # Gather nodes information from yaml file
    mgmt_version = \
        kaas_manager.get_mgmt_cluster().spec['providerSpec']['value']['kaas'][
            'release']
    child_data = yaml.safe_load(
        templates.render_template(settings.BM_CHILD_SETTINGS_YAML_PATH,
                                  options=dict(mgmt_version=mgmt_version)))[0]
    bm_hosts_data = child_data.get('nodes')
    ns.create_bm_statics(child_data)

    controller_scale = []
    for node in bm_hosts_data:
        if 'child-scale-controller' in node.get('si_roles', []):
            cred_name = node['name'] + '-cred'
            # Dirty hack, for not copy-paste bmh_name|cred_name across whole
            # test.
            bmh_name = utils.render_bmh_name(node['name'],
                                             cluster_name,
                                             node.get('bootUEFI', True),
                                             node['bmh_labels'],
                                             si_roles=node.get('si_roles',
                                                               False))
            _node = node.copy()
            _node.update({'bmh_name': bmh_name,
                          'cred_name': cred_name})
            controller_scale.append(_node)
            if not kaas_manager.api.secrets.present(name=cred_name,
                                                    namespace=namespace_name):
                secret_data = {
                    "username": node.get('ipmi').get('username'),
                    "password": node.get('ipmi').get('password')
                }
                ns.create_secret(name=cred_name, data=secret_data,
                                 region=region, provider="baremetal")

    assert len(controller_scale) >= 1, "BMH with si_roles " \
                                       "'child-scale-controller' not found"
    new_nodes_name = []
    for node in controller_scale:
        bmh_name = node['bmh_name']
        new_nodes_name.append(bmh_name)
        ns.create_baremetalhost(bmh_name=bmh_name,
                                bmh_secret=node['cred_name'],
                                bmh_mac=node['networks'][0]['mac'],
                                bmh_ipmi=node['ipmi'],
                                hardwareProfile=node.get('hardwareProfile',
                                                         False),
                                labels=node['bmh_labels'],
                                bootUEFI=node.get('bootUEFI', True))
    LOG.info(f"New BMH nodes:\n{new_nodes_name}")
    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='ready',
                                    retries=20,
                                    interval=60)
    cluster = ns.get_cluster(cluster_name)

    # Create control plane machines
    for node in controller_scale:
        LOG.info('Create BareMetal machine')
        custom_bmhp = False
        if node.get('bmh_profile'):
            custom_bmhp = {
                'namespace': namespace_name,
                'name': node['bmh_profile']
            }
        cluster.create_baremetal_machine(
            genname=node['bmh_name'],
            node_pubkey_name=public_key_name,
            matchlabels={'kaas.mirantis.com/'
                         'baremetalhost-id':
                             node['bmh_labels']
                             ['kaas.mirantis.com/baremetalhost-id']},
            baremetalhostprofile=custom_bmhp,
            l2TemplateSelector=node.get('l2TemplateSelector', dict()),
            labels=node['machine_labels'],
            node_labels=node.get('node_labels', {}),
            distribution=node.get('distribution', None),
        )

    # Wait Baremetal hosts be provisioned
    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='provisioned',
                                    retries=30,
                                    interval=60)

    # Waiting for machines are Ready and helmbundles are deployed
    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()
    cluster.check.check_k8s_nodes()
    cluster.check.check_cluster_readiness()
    cluster.check.check_helmbundles()

    # Collecting artifacts
    cluster.store_k8s_artifacts()

    if cluster.lcm_type_is_ucp:
        # Check/wait for correct docker service replicas in cluster
        cluster.check.check_actual_expected_docker_services()
    cluster.check.check_k8s_pods()
    cluster.check.check_actual_expected_pods()


def test_add_delete_bm_worker_node(kaas_manager, ipmi_client):
    """Delete one worker node."""

    # Get namespace
    LOG.info(f"Namespace name: {namespace_name}")
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)

    # Create ssh key
    public_key_name = "{cluster_name}-key".format(
        cluster_name=cluster_name)
    LOG.info("Public key name - {public_key_name}".format(
        public_key_name=public_key_name))
    ns.get_publickey(public_key_name)

    mgmt_version = \
        kaas_manager.get_mgmt_cluster().spec['providerSpec']['value']['kaas'][
            'release']
    # Gather nodes information from yaml file
    child_data = yaml.safe_load(
        templates.render_template(settings.BM_CHILD_SETTINGS_YAML_PATH,
                                  options=dict(mgmt_version=mgmt_version)))[0]
    ns.create_bm_statics(child_data)
    bm_hosts_data = child_data['nodes']

    worker_scale = []
    for node in bm_hosts_data:
        if 'nodeforscale' in node.get('si_roles', []):
            cred_name = node['name'] + '-cred'
            # Dirty hack, for not copy-paste bmh_name|cred_name across whole
            # test.
            bmh_name = utils.render_bmh_name(node['name'],
                                             cluster_name,
                                             node.get('bootUEFI', True),
                                             node['bmh_labels'],
                                             si_roles=node.get('si_roles',
                                                               False))
            _node = node.copy()
            _node.update({'bmh_name': bmh_name,
                          'cred_name': cred_name})
            worker_scale.append(_node)
            if not kaas_manager.api.secrets.present(name=cred_name,
                                                    namespace=namespace_name):
                secret_data = {
                    "username": node.get('ipmi').get('username'),
                    "password": node.get('ipmi').get('password')
                }
                ns.create_secret(name=cred_name, data=secret_data,
                                 region=region, provider="baremetal")

    assert len(worker_scale) >= 1, "BMH with si_roles " \
                                   "'nodeforscale' not found"

    new_nodes_name = []
    for node in worker_scale:
        bmh_name = node['bmh_name']
        new_nodes_name.append(bmh_name)
        ns.create_baremetalhost(bmh_name=bmh_name,
                                bmh_secret=node['cred_name'],
                                bmh_mac=node['networks'][0]['mac'],
                                bmh_ipmi=node['ipmi'],
                                hardwareProfile=node.get('hardwareProfile',
                                                         False),
                                labels=node['bmh_labels'],
                                bootUEFI=node.get('bootUEFI', True))
    LOG.info(f"New BMH nodes:\n{new_nodes_name}")

    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='ready',
                                    retries=20,
                                    interval=60)

    # Create worker machines
    for node in worker_scale:
        LOG.info('Create BareMetal machine')
        custom_bmhp = False
        if node.get('bmh_profile'):
            custom_bmhp = {
                'namespace': namespace_name,
                'name': node['bmh_profile']
            }
        cluster.create_baremetal_machine(
            genname=node['bmh_name'],
            node_pubkey_name=public_key_name,
            matchlabels={'kaas.mirantis.com/'
                         'baremetalhost-id':
                             node['bmh_labels']
                             ['kaas.mirantis.com/baremetalhost-id']},
            baremetalhostprofile=custom_bmhp,
            l2TemplateSelector=node.get('l2TemplateSelector', dict()),
            labels=node['machine_labels'],
            node_labels=node.get('node_labels', {}),
            distribution=node.get('distribution', None),
        )

    # Wait Baremetal hosts be provisioned
    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='provisioned',
                                    retries=30,
                                    interval=60)

    # Waiting for cluster,pods,hb are Ready after machines were added
    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()
    cluster.check.check_k8s_nodes()
    cluster.check.check_cluster_readiness()
    cluster.check.check_helmbundles()

    cluster.check.check_k8s_pods()

    # Delete machine(s)
    # TODO(alexz) Reference to nodename- not a best choice here. probably
    #  we should add logic into nodeLabels
    machines_for_delete = [machine for machine in ns.get_machines()
                           if 'nodeforscale' in machine.name]
    for machine in machines_for_delete:
        machine.delete()
        cluster.wait_machine_deletion(machine.name)
    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='ready',
                                    retries=30,
                                    interval=60)

    # Waiting for cluster,pods,hb are Ready after machines were deleted
    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()
    cluster.check.check_k8s_nodes()
    cluster.check.check_cluster_readiness()
    cluster.check.check_helmbundles()

    cluster.check.check_k8s_pods()

    # Recreate BMH and check it has been inspected
    bmh_for_delete = [bmh for bmh in ns.get_baremetalhosts()
                      if 'nodeforscale' in bmh.name]
    LOG.info("Recreating BHM(s) and check it has been inspected")
    for bmh in bmh_for_delete:
        bmc_data = cluster.get_bmc_data(bmh)
        bmh_name = bmh.name
        LOG.info(f"Deleting BMH: {bmh_name}")
        bmh.delete()
        ns.wait_baremetalhost_deletion(bmh_name)
        ipmi = ipmi_client.interface(ipmi_host=bmc_data[0],
                                     ipmi_user=bmc_data[1],
                                     ipmi_password=bmc_data[2])
        power_on = ipmi.get_chassis_status().power_on
        LOG.info("Chassis power on status: {}".format(power_on))
        if settings.BM_CHECK_BMC_STATUS_AFTER_DELETE:
            assert not power_on, \
                "Chassis status is not powered off. "\
                "Now power_on status is {}".format(power_on)
    LOG.info(f"Next BMH(s) were deleted successfully\n:{bmh_for_delete}")
    LOG.info("Recreating Baremetal host(s)")
    for node in worker_scale:
        cred_name = node['name'] + '-cred'
        bmh_name = node['bmh_name']
        # Wait for secret deletion after node was deleted
        waiters.wait(lambda: not kaas_manager.api.secrets.present(
            name=cred_name, namespace=namespace_name),
                     timeout_msg=f"Secret {cred_name} "
                                 f"was not deleted succesfully")

        secret_data = {
                "username": node['ipmi']['username'],
                "password": node['ipmi']['password']
        }
        ns.create_secret(name=cred_name, data=secret_data,
                         region=region, provider="baremetal")

        new_nodes_name.append(bmh_name)
        ns.create_baremetalhost(bmh_name=bmh_name,
                                bmh_secret=node['cred_name'],
                                bmh_mac=node['networks'][0]['mac'],
                                bmh_ipmi=node['ipmi'],
                                hardwareProfile=node.get('hardwareProfile',
                                                         False),
                                labels=node['bmh_labels'],
                                bootUEFI=node.get('bootUEFI', True))

    ns.wait_baremetalhosts_statuses(nodes=new_nodes_name,
                                    wait_status='ready',
                                    retries=30,
                                    interval=60)
    LOG.info("Baremetal host(s) is succesfully recreated")

    # Refresh the expected objects stored in the cluster
    cluster._refresh_expected_objects()
    if cluster.lcm_type_is_ucp:
        # Check/wait for correct docker service replicas in cluster
        cluster.check.check_actual_expected_docker_services()
    # Comparing pods moved to last step due to PRODX-2929
    cluster.check.check_actual_expected_pods()
    cluster.check.check_cluster_readiness()
