import exec_helpers
import yaml

from si_tests import logger
from si_tests import settings
from si_tests.utils import waiters, utils

LOG = logger.logger

ms_server_name = 'ms01'
ms_segment_name = "allcomputes"
ms_network_name = "masakari_net"
ms_subnet_name = "masakari_subnet"
ms_subnet_range = "10.1.2.0/24"


def create_masakari_resources(openstack_client_manager,
                              segment, vm):
    LOG.info("1.1 Create Masakari segment")
    openstack_client_manager.segment.create([segment, 'auto',
                                             'compute'])
    hosts = openstack_client_manager.host.list([])

    # Add Hosts to Masakari segment
    LOG.info("1.2 Add Hosts to Masakari segment")
    for hs in hosts:
        if hs.get('Service') == 'compute':
            openstack_client_manager.segment.host_create(
                [hs.get('Host Name'), 'compute', 'SSH', segment])

    # Create a network
    LOG.info("Step 2: Boot VM on compute host")
    LOG.info("2.1: Create network")
    openstack_client_manager.network.create([ms_network_name])

    # Create a subnet
    LOG.info("2.2: Create subnetwork")
    openstack_client_manager.subnet.create(["--subnet-range",
                                            ms_subnet_range, "--network",
                                            ms_network_name, ms_subnet_name])

    # Create a VM
    LOG.info("2.3: Boot VM on compute host")
    openstack_client_manager.server.create(
        ["--image=Cirros-5.1", "--flavor", "m1.extra_tiny_test", "--network",
         ms_network_name, "--property", "HA_Enabled=True", vm])

    # Check VM is running
    LOG.info("2.4: Check VM is running")
    waiters.wait(lambda: openstack_client_manager.server.show(
        [ms_server_name])["OS-EXT-STS:power_state"] in ["Running", 1],
                 timeout=60, interval=10)


def ssh_node(hostname):
    with open(settings.NODES_INFO) as f:
        dictionary = yaml.safe_load(f)
    private_key = dictionary.get(hostname, {}).get('ssh')[
        'private_key']
    assert private_key, "Private key is not defined"
    pkey = utils.get_rsa_key(private_key)
    node_ip = dictionary.get(hostname, {}).get('ip')['address']
    username = dictionary.get(hostname, {}).get('ssh')['username']
    auth = exec_helpers.SSHAuth(username=username,
                                password='', key=pkey)
    ssh = exec_helpers.SSHClient(
        host=node_ip, port=22, auth=auth)
    ssh.logger.addHandler(logger.console)
    ssh.sudo_mode = True
    return ssh


def wait_masakari_notification(openstack_client_manager, host_uuid,
                               notification_type, status="finished"):
    waiters.wait(lambda: openstack_client_manager.notification.list([])[0][
                             "source_host_uuid"] == host_uuid and
                 openstack_client_manager.notification.list([])[0]["type"] ==
                 notification_type, timeout=180, interval=30)
    waiters.wait(lambda: openstack_client_manager.notification.list([])[0][
                             "status"] == status,
                 timeout=300, interval=30)


def test_host_monitor(openstack_client_manager):
    """Verify Masakari host monitor.
    https://mirantis.testrail.com/index.php?/cases/view/4963423

     Scenario:
         1. Create Masakari resources
         2. Boot VM on compute host
         3. Simulate host crush
         4. Wait k8s node is not ready
         5. Check Masakari notification
         6. Check that compute service is disabled on crashed compute
         7. Check VM host is changed
         8. Clean up

     """

    try:
        # Create Masakari segment
        LOG.info("Step 1: Create masakari resources")
        create_masakari_resources(openstack_client_manager,
                                  segment=ms_segment_name,
                                  vm=ms_server_name)

        # get values of the created Masakari resources
        server = openstack_client_manager.server.show([ms_server_name])
        hostname = server["OS-EXT-SRV-ATTR:host"]
        segment_hosts = openstack_client_manager.segment.host_list(
            [ms_segment_name])
        host_uuid = [x["uuid"] for x in segment_hosts if x["name"] ==
                     hostname][0]
        hyper_indx = None
        for hv in openstack_client_manager.hypervisor.list([]):
            if hostname in hv["Hypervisor Hostname"]:
                hyper_indx = openstack_client_manager.hypervisor.list(
                    []).index(hv)
                break

        # Simulate host crush on the node where the instance is running
        LOG.info("Step 3: Simulate host crush")
        ssh = ssh_node(hostname)
        ssh.check_call("sudo systemctl stop docker")

        # Wait k8s node is not ready anymore
        LOG.info("Step 4: Wait k8s node is not ready anymore")
        waiters.wait(lambda: hostname not in [x["Host Name"] for x
                     in openstack_client_manager.host.list([])], timeout=300,
                     interval=30)

        # Check Masakari notification
        LOG.info("Step 5: Check Masakari notification")
        wait_masakari_notification(openstack_client_manager, host_uuid,
                                   notification_type="COMPUTE_HOST")

        # Check that compute service is disabled on crashed compute
        LOG.info("Step 6: Check that compute service is disabled on crashed "
                 "compute")
        assert hostname in [x["Host"] for x in
                            openstack_client_manager.compute_service.list([])
                            if (x["Host"] == hostname
                                and x["Status"] == "disabled")],\
            "compute service is NOT disabled on crashed compute"

        # Check VM's host is changed
        LOG.info("Step 7: Check VM's host is changed")
        assert hostname != openstack_client_manager.server.show(
            [ms_server_name])["OS-EXT-SRV-ATTR:host"], \
            "VM's host is NOT changed"
        assert openstack_client_manager.server.show(
            [ms_server_name])["OS-EXT-STS:power_state"] in ["Running", 1], \
            "VM's instance is NOT running"

    finally:
        # Restore Node and cleanup
        LOG.info("Step 8: Restore Node and cleanup")
        LOG.info("8.1: Delete VM")
        openstack_client_manager.server.delete([ms_server_name])
        LOG.info("8.2: Wait VM is not present anymore")
        waiters.wait(lambda: ms_server_name not in [x["Name"] for x in
                     openstack_client_manager.server.list([])], timeout=60,
                     interval=10)
        LOG.info("8.3: Delete Masakari segment")
        openstack_client_manager.segment.delete([ms_segment_name])
        LOG.info("8.4: Delete Masakari net")
        openstack_client_manager.network.delete([ms_network_name])
        LOG.info("8.5: Delete Masakari subnet")
        openstack_client_manager.subnet.delete([ms_subnet_name])
        LOG.info("8.6: Start docker on node")
        ssh.check_call("sudo systemctl start docker")
        LOG.info("8.7: Wait k8s node is ready again")
        waiters.wait(lambda: openstack_client_manager.hypervisor.list([])[
                                 hyper_indx]
                     ["State"] == "up", timeout=300, interval=30)
        LOG.info("8.8: Restore compute service on the crushed node")
        openstack_client_manager.compute_service.set(
            ["--enable", hostname, "nova-compute"])


def test_instance_monitor(openstack_client_manager):
    """Verify Masakari instance monitor.
    https://mirantis.testrail.com/index.php?/cases/view/4963424

     Scenario:
         1. Create Masakari resources
         2. Boot VM on compute host
         3. Simulate instance crush
         4. Check Masakari notification
         5. Check VM instance is running
         6. Clean up

     """
    try:
        # Create Masakari segment
        LOG.info("Step 1: Create masakari resources")
        create_masakari_resources(openstack_client_manager,
                                  segment=ms_segment_name,
                                  vm=ms_server_name)

        # get values of the created Masakari resources
        server = openstack_client_manager.server.show([ms_server_name])
        hostname = server["OS-EXT-SRV-ATTR:host"]
        segment_hosts = openstack_client_manager.segment.host_list(
            [ms_segment_name])
        host_uuid = [x["uuid"] for x in segment_hosts if x["name"] ==
                     hostname][0]

        # Simulate instance crush
        LOG.info("Step 3: Simulate instance crush")
        ssh = ssh_node(hostname)
        ssh.check_call("sudo kill -9 $(pidof qemu-system-x86_64)")

        # Check Masakari notification
        LOG.info("Step 4: Check Masakari notification")
        wait_masakari_notification(openstack_client_manager, host_uuid,
                                   notification_type="VM")

        # Wait VM is running again
        LOG.info("Step 5: Wait VM is running again")
        waiters.wait(lambda: openstack_client_manager.server.show(
            [ms_server_name])["OS-EXT-STS:power_state"] in ["Running", 1],
                     timeout=180, interval=30)

    finally:
        # Restore Node and cleanup
        LOG.info("Step 6: Restore Node and cleanup")
        LOG.info("6.1: Delete VM")
        openstack_client_manager.server.delete([ms_server_name])
        LOG.info("6.2: Wait VM is not present anymore")
        waiters.wait(lambda: ms_server_name not in [x["Name"] for x in
                     openstack_client_manager.server.list([])], timeout=60,
                     interval=10)
        LOG.info("6.3: Delete Masakari segment")
        openstack_client_manager.segment.delete([ms_segment_name])
        LOG.info("6.4: Delete Masakari net")
        openstack_client_manager.network.delete([ms_network_name])
        LOG.info("6.5: Delete Masakari subnet")
        openstack_client_manager.subnet.delete([ms_subnet_name])
